use bevy_ecs::{entity::Entity, prelude::Component};

use super::Scalar;

#[derive(Clone, Copy, PartialEq, Eq, Debug)]
pub enum VarStatus {
    Free,
    Dependent,
    Unique,
    Overconstrained,
}

impl VarStatus {
    pub fn merge(self, other: Self) -> Self {
        use VarStatus::*;
        match (self, other) {
            (Free, Free) => Free,
            (Unique, Unique) => Unique,
            (Overconstrained, _) | (_, Overconstrained) => Overconstrained,
            _ => Dependent,
        }
    }
}

#[derive(Clone, Copy, Debug, Component)]
pub struct Var<T = Scalar> {
    pub value: T,
    pub status: VarStatus,
}

impl<T: Default> Default for Var<T> {
    fn default() -> Self {
        Self::new_free(T::default())
    }
}

impl<T> Var<T> {
    pub fn new_free(value: T) -> Self {
        Self {
            value,
            status: VarStatus::Free,
        }
    }

    pub(super) fn merge<U, V, F>(self, other: Var<U>, by: F) -> Var<V>
    where
        F: FnOnce(T, U) -> V,
    {
        Var {
            value: by(self.value, other.value),
            status: self.status.merge(other.status),
        }
    }
}

impl<T> std::ops::Deref for Var<T> {
    type Target = T;

    fn deref(&self) -> &Self::Target {
        &self.value
    }
}

impl<T> std::ops::DerefMut for Var<T> {
    fn deref_mut(&mut self) -> &mut Self::Target {
        &mut self.value
    }
}

pub type VarId = Entity;
