use nalgebra::SMatrix;

use crate::geometry::Scalar;

pub trait WriteResiduals<Index> {
    fn write(&self, index: Index, residuals: &mut [Scalar]);
}

impl<I> WriteResiduals<I> for Scalar
where
    usize: From<I>,
{
    fn write(&self, index: I, residuals: &mut [Scalar]) {
        residuals[usize::from(index)] = *self;
    }
}

impl<I, const R: usize, const C: usize> WriteResiduals<I> for SMatrix<Scalar, R, C>
where
    usize: From<I>,
{
    fn write(&self, index: I, residuals: &mut [Scalar]) {
        residuals[usize::from(index)..][..R * C].copy_from_slice(self.as_slice());
    }
}

impl<T1, T2, I1, I2> WriteResiduals<(I1, I2)> for (T1, T2)
where
    T1: WriteResiduals<I1>,
    T2: WriteResiduals<I2>,
{
    fn write(&self, index: (I1, I2), residuals: &mut [Scalar]) {
        self.0.write(index.0, residuals);
        self.1.write(index.1, residuals);
    }
}

impl<T1, T2, T3, I1, I2, I3> WriteResiduals<(I1, I2, I3)> for (T1, T2, T3)
where
    T1: WriteResiduals<I1>,
    T2: WriteResiduals<I2>,
    T3: WriteResiduals<I3>,
{
    fn write(&self, index: (I1, I2, I3), residuals: &mut [Scalar]) {
        self.0.write(index.0, residuals);
        self.1.write(index.1, residuals);
        self.2.write(index.2, residuals);
    }
}

impl<T1, T2, T3, T4, I1, I2, I3, I4> WriteResiduals<(I1, I2, I3, I4)> for (T1, T2, T3, T4)
where
    T1: WriteResiduals<I1>,
    T2: WriteResiduals<I2>,
    T3: WriteResiduals<I3>,
    T4: WriteResiduals<I4>,
{
    fn write(&self, index: (I1, I2, I3, I4), residuals: &mut [Scalar]) {
        self.0.write(index.0, residuals);
        self.1.write(index.1, residuals);
        self.2.write(index.2, residuals);
        self.3.write(index.3, residuals);
    }
}
